/*
   This file is part of the index package,
   Copyright (C) 1997 Harold Tay, haroldtay@pacific.net.sg

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA
 */
/* make an index.
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <gdbm.h>
#include <stdarg.h>
#include <sys/stat.h>
#include "hash.h"
#include "index.h"
#include "codec.h"

extern int string_init(void);
extern char * string_dup(char * text);
extern void string_free(void);

static char *
getline(void)
{
  static char * buf = NULL;
  static int size;
  int len = 0;

  if( NULL == buf )
    buf = malloc(size = 1024);

  do{
    if( NULL == fgets(buf + len, size - len, stdin) )
      return(NULL);
    len = strlen(buf);
    if( len >= size )
      buf = realloc(buf, size *= 2);
  }while( '\n' != buf[len - 1] );

  buf[len - 1] = '\0';

  return(buf);
}

/* LOCS_APPEAR_ONCE means that the same location cannot appear on
   the input multiple times.  This is true if we are being fed with
   a "normal" scanner, which works on one location (file) at a time,
   then moves on to another location (file), never to return to the
   previous ones.

   If LOCS_APPEAR_ONCE, then we do not need a locations hash table,
   and we can save some memory that way, and also because we don't
   have to save the string.
 */

int words = -1;
int locations_size = 0;

#ifndef LOCS_APPEAR_ONCE

int locations = -1;

static int
getlocoff(char * location)
{
  int * d;

  d = (int *)hash_fetch(locations, location);
  if( -1 == (int)d ){
    int locoff;
    locoff = locations_size;
    locations_size += write(1, location, strlen(location) + 1);
    d = malloc(sizeof(int));
    *d = locoff;
    hash_insert(locations, string_dup(location), (char *)d);
    return(locoff);
  }else
    return(*d);
}

#else

static int
getlocoff(char * location)
{
  static char * prevloc = NULL;
  static int prevoff = 0;

  if( NULL == prevloc || strcmp(prevloc, location) ){
    int locoff;
    if( prevloc )free(prevloc);
    prevloc = strdup(location);
    prevoff = locoff = locations_size;
    locations_size += write(1, location, strlen(location) + 1);
    return(locoff);
  }else
    return(prevoff);
}
#endif /* LOCS_APPEAR_ONCE */

void
storeit(char * word, char * location)
{
  int * d;
  int locoff;

  /* get offset of location */

  locoff = getlocoff(location);

  /* with locations offset, we can insert the word */

  d = (int *)hash_fetch(words, word);
  if( -1 == (int)d ){
    char * tmp;
    d = malloc(2*sizeof(int));
    d[0] = 1;
    d[1] = locoff;
    tmp = string_dup(word);
    hash_insert(words, tmp, (char *)d);
  }else{
    int i;
    int len = *d;
    for(i = 1; i <= len; i++)
      if( d[i] == locoff )return;
    if( *d & 0x1 ){
      /* odd lengths are exact, even lengths have extra space
         so we malloc only every other time.  This means the
         malloced area is always aligned to 2 ints (which is probably
         what malloc does anyway) */
      int * r = malloc((len + 3)*sizeof(int));
      memcpy(r, d, (len + 1)*sizeof(int));
      r[0]++;
      r[len + 1] = locoff;
      d = (int *)hash_replace(words, word, (char *)r);
      if( d )free(d);
    }else{
      d[0]++;
      d[len + 1] = locoff;
    }
  }
}

/* dump() will be called for each word we encountered. */

unsigned words_size = 0;

static int
dump(char * w, char * dc)
{
  int * ary;
  int arlen;
  int wlen;
  char * buf;

  ary = ((int *)dc)+1;
  arlen = *(int *)dc;
  wlen = strlen(w);

  /* write the word with trailing nul */
  words_size += write(1, w, wlen + 1);

  /* Create and write the compressed buffer */
  {
    int size = codec_co(ary, arlen, &buf);
    words_size += write(1, buf, size);
  }

  return(0);
}

void
fail(char * fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, "\n");
  exit(1);
}

void
init(void)
{
  string_init();
#ifndef LOCS_APPEAR_ONCE
  locations = hash_create(NULL, NULL, 2048);
  if( -1 == locations )
    fail("Can't create hash table for locations: %s?", strerror(errno));
#endif
  words = hash_create(NULL, NULL, 2048);
  if( -1 == words )
    fail("Can't create hash table for words: %s?", strerror(errno));
}

static int
freeit(char * cp, char * dc)
{
  free(dc);
  return(0);
}

void
cleanup(void)
{
  hash_walk(words, freeit);
  hash_destroy(words);
#ifndef LOCS_APPEAR_ONCE
  hash_walk(locations, freeit);
  hash_destroy(locations);
#endif
  string_free();
}

struct meta meta = {MAGIC, 0, };

int
main(int argc, char ** argv)
{
  char * word;
  char * line;
  int nr;
  char * location = NULL;

  if( argc > 1 ){
    fprintf(stderr, "Usage: %s [<filename] [>index]\n"
      "Input is always from standard in,\n"
      "Output is always to standard out.\n", argv[0]);
    exit(1);
  }

  init();
  atexit(cleanup);

  /* allocate some space at start of file for meta */
  write(1, &meta, sizeof(meta));

  for(nr = 0; (line = getline()); nr++){
    char * cp;

    if( *line == '\001' ){
      if( location )free(location);
      location = strdup(line + 1);
      if( (cp = strchr(location, '\n')) )*cp = '\0';
      continue;
    }

    word = line;
    if( (cp = strchr(word, '\n')) )*cp = '\0';

    storeit(word, location);
  }

  /* this will write the words */
  hash_walk(words, dump);

  /* Now go back and write the meta information */
  lseek(1, 0, SEEK_SET);
  meta.locations_size = locations_size;
  meta.words_size = words_size;
  write(1, &meta, sizeof(meta));
  exit(0);
}
